"""A creation workflow that adds an existing Layer 2 Circuit to the database."""

from typing import Any, Self
from uuid import uuid4

from orchestrator import step, workflow
from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, SubscriptionLifecycle
from orchestrator.workflow import StepList, begin, done
from orchestrator.workflows.steps import resync, set_status, store_process_subscription
from pydantic import BaseModel, ConfigDict, model_validator
from pydantic_forms.types import UUIDstr

from gso.products import ProductName
from gso.products.product_blocks.layer_2_circuit import Layer2CircuitSideBlockInactive, Layer2CircuitType
from gso.products.product_blocks.service_binding_port import ServiceBindingPortInactive
from gso.products.product_types.edge_port import EdgePort
from gso.products.product_types.layer_2_circuit import (
    ImportedLayer2CircuitInactive,
    Layer2CircuitServiceType,
)
from gso.services.partners import get_partner_by_name
from gso.services.subscriptions import get_product_id_by_name
from gso.utils.shared_enums import SBPType
from gso.utils.types.interfaces import BandwidthString
from gso.utils.types.virtual_identifiers import VC_ID, VLAN_ID


def initial_input_form_generator() -> FormGenerator:
    """Generate a form that can be pre-filled using an :term:`API` endpoint."""

    class ServiceBindingPortInput(BaseModel):
        edge_port: UUIDstr
        vlan_id: VLAN_ID

    class ImportLayer2CircuitForm(FormPage):
        model_config = ConfigDict(title="Import Layer 2 Circuit")

        service_type: Layer2CircuitServiceType
        partner: str
        geant_sid: str
        vc_id: VC_ID
        layer_2_circuit_side_a: ServiceBindingPortInput
        layer_2_circuit_side_b: ServiceBindingPortInput
        layer_2_circuit_type: Layer2CircuitType
        vlan_range_lower_bound: VLAN_ID | None = None
        vlan_range_upper_bound: VLAN_ID | None = None
        policer_enabled: bool = False
        policer_bandwidth: BandwidthString | None = None
        policer_burst_rate: BandwidthString | None = None

        @model_validator(mode="after")
        def tagged_layer_2_circuit_has_vlan_bounds(self) -> Self:
            """If a Layer 2 Circuit is tagged, it must have a :term:`VLAN` range set."""
            if self.layer_2_circuit_type == Layer2CircuitType.TAGGED and (
                self.vlan_range_lower_bound is None or self.vlan_range_upper_bound is None
            ):
                msg = (
                    f"A tagged Layer 2 Circuit must have a VLAN range set. Received lower: "
                    f"{self.vlan_range_lower_bound}, upper: {self.vlan_range_upper_bound}."
                )
                raise ValueError(msg)
            return self

        @model_validator(mode="after")
        def policer_bandwidth_required_if_policer_enabled(self) -> Self:
            """If the policer is enabled, the bandwidth and burst rate must be set."""
            if self.policer_enabled and (self.policer_bandwidth is None or self.policer_burst_rate is None):
                msg = (
                    f"If the policer is enabled, the bandwidth and burst rate must be set. Received bandwidth: "
                    f"{self.policer_bandwidth}, burst rate: {self.policer_burst_rate}."
                )
                raise ValueError(msg)
            return self

    user_input = yield ImportLayer2CircuitForm
    return user_input.model_dump()


@step("Create subscription")
def create_subscription(partner: str, service_type: Layer2CircuitServiceType) -> State:
    """Create a new subscription object."""
    partner_id = get_partner_by_name(partner)["partner_id"]
    product_id = get_product_id_by_name(ProductName(service_type))
    subscription = ImportedLayer2CircuitInactive.from_product_id(product_id, partner_id)

    return {"subscription": subscription, "subscription_id": subscription.subscription_id}


@step("Initialize subscription")
def initialize_subscription(
    subscription: ImportedLayer2CircuitInactive,
    geant_sid: str,
    layer_2_circuit_side_a: dict[str, Any],
    layer_2_circuit_side_b: dict[str, Any],
    layer_2_circuit_type: Layer2CircuitType,
    vlan_range_lower_bound: VLAN_ID | None,
    vlan_range_upper_bound: VLAN_ID | None,
    policer_enabled: bool,  # noqa: FBT001
    policer_bandwidth: BandwidthString | None,
    policer_burst_rate: BandwidthString | None,
    vc_id: VC_ID,
) -> State:
    """Initialize the subscription object."""
    layer_2_circuit_sides = []
    for circuit_side_data in [layer_2_circuit_side_a, layer_2_circuit_side_b]:
        sbp = ServiceBindingPortInactive.new(
            uuid4(),
            edge_port=EdgePort.from_subscription(subscription_id=circuit_side_data["edge_port"]).edge_port,
            sbp_type=SBPType.L2,
            vlan_id=circuit_side_data["vlan_id"],
            geant_sid=geant_sid,
            is_tagged=layer_2_circuit_type == Layer2CircuitType.TAGGED,
            custom_firewall_filters=False,
        )
        layer2_circuit_side = Layer2CircuitSideBlockInactive.new(uuid4(), sbp=sbp)
        layer_2_circuit_sides.append(layer2_circuit_side)
    subscription.layer_2_circuit.layer_2_circuit_sides = layer_2_circuit_sides
    subscription.layer_2_circuit.virtual_circuit_id = vc_id
    subscription.layer_2_circuit.layer_2_circuit_type = layer_2_circuit_type
    subscription.layer_2_circuit.vlan_range_lower_bound = vlan_range_lower_bound
    subscription.layer_2_circuit.vlan_range_upper_bound = vlan_range_upper_bound
    subscription.layer_2_circuit.policer_enabled = policer_enabled
    subscription.layer_2_circuit.bandwidth = policer_bandwidth
    subscription.layer_2_circuit.policer_burst_rate = policer_burst_rate
    subscription.description = f"{subscription.product.name} - {subscription.layer_2_circuit.virtual_circuit_id}"

    return {"subscription": subscription}


@workflow(
    "Create imported Layer 2 Circuit",
    initial_input_form=initial_input_form_generator,
    target=Target.CREATE,
)
def create_imported_layer_2_circuit() -> StepList:
    """Import a Layer 2 Circuit without provisioning it."""
    return (
        begin
        >> create_subscription
        >> store_process_subscription(Target.CREATE)
        >> initialize_subscription
        >> set_status(SubscriptionLifecycle.ACTIVE)
        >> resync
        >> done
    )
